// Copyright (C) 2007 Damien Katz <damien_katz@yahoo.com>.

// This file is licenced under a Creative Commons Attribution 3.0 Unported
// Licence <http://creativecommons.org/licenses/by/3.0/>.

/*
 * Page class for browse/index.html
 */
function CouchIndexPage() {
  page = this;

  this.addDatabase = function() {
    var dbName = prompt("Please enter the name of the database:");
    if (dbName) {
      try {
        new CouchDB(dbName).createDb();
      } catch (e) {
        alert(e.error.reason);
        return;
      }
      if (window !== parent) parent.setTimeout("updateDatabaseList()", 500);
      window.open("database.html?" + dbName, "content");
    }
    return false;
  }

  this.updateDatabaseListing = function() {
    var allDbs = CouchDB.allDbs();
    for (var i = 0; i < allDbs.length; i++) {
      var dbName = allDbs[i];
      var info = new CouchDB(dbName).info();
      $("#databases tbody.content").append(
        "<tr><th><a href='database.html?" + dbName + "'>" +
        dbName + "</a></th><td>" + info.doc_count +"</td><td>" +
        info.update_seq + "</td></tr>");
      $("#databases tbody tr:odd").addClass("odd");
      $("#databases tbody.footer tr td").text(allDbs.length + " database(s)");
    }
  }

}

/*
 * Page class for browse/database.html
 */
function CouchDatabasePage() {
  var urlParts = location.search.substr(1).split("/");
  var dbName = urlParts.shift();
  var viewName = (urlParts.length > 0) ? urlParts.join("/") : "";
  var db = new CouchDB(dbName);

  this.dbName = dbName;
  this.viewName = viewName;
  this.db = db;
  page = this;

  this.addDocument = function() {
    var docId = prompt("Please enter the ID of the document, or leave " +
                       "empty to use an auto-generated ID:");
    if (docId !== null) {
      var result = db.save(docId ? {_id: docId} : {});
      location.href = "document.html?" + dbName + "/" + result.id;
    }
  }

  this.deleteDatabase = function() {
    if (confirm("Are you sure you want to delete this database?")) {
      db.deleteDb();
      location.href = "index.html";
      if (window !== null) {
        parent.$("#dbs li").filter(function(index) {
          return $("a", this).text() == dbName;
        }).remove();
      }
    }
  }

  this.populateViewsMenu = function() {
    var designDocs = db.allDocs({
      startkey: "_design/",
      endkey: "_design/ZZZ"
    });
    $("#switch select").each(function() {
      this.options.length = 2;
      for (var i = 0; i < designDocs.rows.length; i++) {
        var doc = db.open(designDocs.rows[i].id);
        var optGroup = $("<optgroup></optgroup>").attr("label", doc._id.substr(8));
        for (var name in doc.views) {
          if (!doc.views.hasOwnProperty(name)) continue;
          $("<option></option>").attr("value", doc._id + "/" + name).text(name)
            .appendTo(optGroup);
        }
        optGroup.appendTo(this);
      }
      this.autocomplete = false;
    });
  }

  this.updateDocumentListing = function(options) {
    if (options === undefined) {
      options = {count: 20};
    }
    $("#paging a").unbind();
    $("#documents tbody.content").empty();

    var result = null;
    if (!viewName) {
      $("#switch select").get(0).selectedIndex = 0;
      result = db.allDocs(options);
    } else {
      $("#switch select").each(function() {
        for (var i = 0; i < this.options.length; i++) {
          if (this.options[i].value == viewName) {
            this.selectedIndex = i;
            break;
          }
        }
      });
      docs = [];
      if (viewName == "_temp_view") {
        $("#tempview").show();
        var query = $("#tempview textarea").val();
        try {
          result = db.query(query, options);
        } catch (e) {
          alert(e.error ? e.error.reason : e);
          return;
        }
      } else {
        result = db.view(viewName.substr(8), options);
      }
    }
    if (result.offset === undefined) {
      result.offset = 0;
    }
    var docs = result.rows;
    if (result.total_rows > 20) {
      if (result.offset > 0) {
        $("#paging a.prev").attr("href", "?offset=" + (result.offset - 20)).click(function() {
          var firstDoc = docs[0];
          page.updateDocumentListing({
            startkey: firstDoc.key !== undefined ? firstDoc.key : null,
            startkey_docid: firstDoc.id,
            skip: 1,
            count: -20
          });
          return false;
        });
      } else {
        $("#paging a.prev").removeAttr("href");
      }
      if (result.total_rows - result.offset > 20) {
        $("#paging a.next").attr("href", "?offset=" + (result.offset + 20)).click(function() {
          var lastDoc = docs[docs.length - 1];
          page.updateDocumentListing({
            startkey: lastDoc.key !== undefined ? lastDoc.key : null,
            startkey_docid: lastDoc.id,
            skip: 1,
            count: 20
          });
          return false;
        });
      } else {
        $("#paging a.next").removeAttr("href");
      }
    }

    for (var i = 0; i < docs.length; i++) {
      var doc = docs[i];
      var row = $("<tr><th><a href='document.html?" +
        dbName + "/" + doc.id + "'>" + doc.id +
        "</a></th></tr>");
      for (prop in doc) {
        if (!doc.hasOwnProperty(prop)) continue;
        if (prop == "id") continue;
        var value = doc[prop];
        $("<td class='extra'></td>").text(
          value !== null ? prettyPrintJSON(value, 0, "") : "null"
        ).appendTo(row);
      }
      row.appendTo("#documents tbody.content");
    }

    $("#documents tbody tr:odd").addClass("odd");
    $("#documents tbody.footer td span").text(
      "Showing " + Math.min(result.total_rows, result.offset + 1) + "-" + (result.offset + docs.length) +
      " of " + result.total_rows +
      " document" + (result.total_rows != 1 ? "s" : ""));
  }

}

/*
 * Page class for browse/database.html
 */
function CouchDocumentPage() {
  var urlParts = location.search.substr(1).split("/");
  var dbName = urlParts.shift();
  var idParts = urlParts.join("/").split("@", 2);
  var docId = idParts[0];
  var docRev = (idParts.length > 1) ? idParts[1] : null;
  var db = new CouchDB(dbName);
  var doc = db.open(docId, {revs_info: true});
  var revs = doc._revs_info;
  delete doc._revs_info;
  if (docRev != null) {
    var result = db.open(docId, {rev: docRev})._doc_revs[0];
    if (result.missing) {
      alert("The requested revision was not found. " +
            "You will be redirected back to the latest revision.");
      location.href = "?" + dbName + "/" + docId;
      return;
    }
    doc = result.ok;
  }

  this.dbName = dbName;
  this.db = db;
  this.doc = doc;
  this.isDirty = false;
  page = this;

  this.addField = function() {
    var fieldName = prompt("Please enter the name of the field:");
    if (fieldName) {
      if (doc[fieldName] !== undefined) {
        alert("The document already contains a field with that name!");
        return;
      }
      doc[fieldName] = null;
      _addRowForField(fieldName);
      page.isDirty = true;
    }
  }

  this.updateFieldListing = function() {
    $("#fields tbody.content").empty();
    for (var prop in doc) {
      if (!doc.hasOwnProperty(prop)) continue;
      _addRowForField(prop);
    }
    if (revs.length > 1) {
      var currentIndex = 0;
      for (var i = 0; i < revs.length; i++) {
        if (revs[i].rev == doc._rev) {
          currentIndex = i;
          break;
        }
      }
      if (currentIndex < revs.length - 1) {
        var prevRev = revs[currentIndex + 1].rev;
        $("#paging a.prev").attr("href", "?" + dbName + "/" + docId + "@" + prevRev);
      }
      if (currentIndex > 0) {
        var nextRev = revs[currentIndex - 1].rev;
        $("#paging a.next").attr("href", "?" + dbName + "/" + docId + "@" + nextRev);
      }
      $("#fields tbody.footer td span").text("Showing revision " +
        (revs.length - currentIndex) + " of " + revs.length);
    }
  }

  this.deleteDocument = function() {
    if (confirm("Are you sure you want to delete this document?")) {
      db.deleteDoc(doc);
      location.href = "database.html?" + dbName;
    }
  }

  this.saveDocument = function() {
    try {
      db.save(doc);
    } catch (e) {
      alert(e.error.reason);
      return;
    }
    page.isDirty = false;
    location.href = "?" + dbName + "/" + docId;
  }

  window.onbeforeunload = function() {
    if (page.isDirty) {
      return "You've made changes to this document that have not been " +
             "saved yet.";
    }
  }

  function _addRowForField(fieldName) {
    var value = _renderValue(doc[fieldName]);
    var row = $("<tr><th></th><td></td></tr>")
      .find("th").append($("<b></b>").text(fieldName)).dblclick(function() {
        _editKey(this, $(this).text());
      }).end()
      .find("td").append(value).dblclick(function() {
        _editValue(this, $(this).prev("th").text());
      }).end()
      .appendTo("#fields tbody.content");
    if (fieldName != "_id" && fieldName != "_rev") {
      row.find("th, td").attr("title", "Double click to edit");
      _initKey(row, fieldName);
      _initValue(value);
    }
    $("#fields tbody tr").removeClass("odd").filter(":odd").addClass("odd");
  }

  function _editKey(cell, fieldName) {
    if (fieldName == "_id" || fieldName == "_rev") return;
    var th = $(cell);
    th.empty();
    var input = $("<input type='text' spellcheck='false'>");
    input.dblclick(function() { return false; }).keypress(function(evt) {
      switch (evt.keyCode) {
        case 13: applyChange(); break;
        case 27: cancelChange(); break;
      }
    });
    var tools = $("<div class='tools'></div>");
    function applyChange() {
      input.nextAll().remove();
      var newName = input.val();
      if (!newName.length || newName == fieldName) {
        cancelChange();
        return;
      }
      doc[newName] = doc[fieldName];
      delete doc[fieldName];
      th.children().remove();
      th.append($("<b></b>").text(newName));
      _initKey(th.parent("tr"), fieldName);
      page.isDirty = true;
    }
    function cancelChange() {
      th.children().remove();
      th.append($("<b></b>").text(fieldName));
      _initKey(th.parent("tr"), fieldName);
    }

    $("<button type='button' class='apply'></button>").click(function() {
      applyChange();
    }).appendTo(tools);
    $("<button type='button' class='cancel'></button>").click(function() {
      cancelChange();
    }).appendTo(tools);
    tools.appendTo(th);
    input.val(fieldName).appendTo(th);
    input.each(function() { this.focus(); this.select(); });
  }

  function _editValue(cell, fieldName) {
    if (fieldName == "_id" || fieldName == "_rev") return;
    var td = $(cell);
    var value = doc[fieldName];
    var needsTextarea = $("dl", td).length > 0;
    td.empty();
    if (needsTextarea) {
      var input = $("<textarea rows='8' cols='40' spellcheck='false'></textarea>");
    } else {
      var input = $("<input type='text' spellcheck='false'>");
    }
    input.dblclick(function() { return false; }).keypress(function(evt) {
      switch (evt.keyCode) {
        case 13:
          if (!needsTextarea) {
            applyChange();
          }
          break;
        case 27:
          cancelChange();
          break;
      }
    });
    var tools = $("<div class='tools'></div>");
    function applyChange() {
      input.nextAll().remove();
      try {
        var newValue = input.val() || "null";
        if (newValue == doc[fieldName]) {
          cancelChange();
          return;
        }
        doc[fieldName] = newValue.parseJSON();
        td.children().remove();
        page.isDirty = true;
        var value = _renderValue(doc[fieldName]);
        td.append(value);
        _initValue(value);
      } catch (err) {
        input.addClass("invalid");
        $("<div class='error'></div>").text(err.toString()).insertAfter(input);
      }
    }
    function cancelChange() {
      td.children().remove();
      var value = _renderValue(doc[fieldName]);
      td.append(value);
      _initValue(value);
    }

    $("<button type='button' class='apply'></button>").click(function() {
      applyChange();
    }).appendTo(tools);
    $("<button type='button' class='cancel'></button>").click(function() {
      cancelChange();
    }).appendTo(tools);
    tools.appendTo(td);
    input.val(prettyPrintJSON(value)).appendTo(td);
    input.each(function() { this.focus(); this.select(); });
  }

  function _initKey(row, fieldName) {
    if (fieldName != "_id" && fieldName != "_rev") {
      $("<button type='button' class='delete' title='Delete field'></button>").click(function() {
        delete doc[fieldName];
        row.remove();
        page.isDirty = true;
        $("#fields tbody tr").removeClass("odd").filter(":odd").addClass("odd");
      }).prependTo(row.find("th"));
    }
  }

  function _initValue(value) {
    value.find("dd").filter(":has(dl)").hide().prev("dt").addClass("collapsed");
    value.find("dd").not(":has(dl)").addClass("inline").prev().addClass("inline");
    value.find("dt.collapsed").click(function() {
      $(this).toggleClass("collapsed").next().toggle();
    });
  }

  function _renderValue(value) {
    var type = typeof(value);
    if (type == "object" && value !== null) {
      var list = $("<dl></dl>");
      for (var i in value) {
        if (!value.hasOwnProperty(i)) continue;
        $("<dt></dt>").text(i).appendTo(list);
        $("<dd></dd>").append(_renderValue(value[i])).appendTo(list);
      }
      return list;
    } else {
      return $("<code></code>").addClass(type).text(
        value !== null ? value.toJSONString() : "null"
      );
    }
  }

}
