// Copyright (C) 2007 Damien Katz <damien_katz@yahoo.com>.

// This file is licenced under a Creative Commons Attribution 3.0 Unported
// Licence <http://creativecommons.org/licenses/by/3.0/>.

/* JSON pretty printing */

function prettyPrintJSON(val, indent, linesep, depth) {
  indent = indent != null ? indent : 4;
  linesep = linesep != null ? linesep : "\n";
  depth = depth != null ? depth : 1;
  var propsep = linesep.length ? "," + linesep : ", ";
  var tab = [];
  for (var i = 0; i < indent * depth; i++) tab.push("");
  tab = tab.join(" ");
  switch (typeof val) {
    case "boolean":
    case "number":
    case "string":
      return val.toJSONString();
    case "object": {
      if (val === null) return "null";
      if (val.constructor == Date) return val.toJSONString();
      var buf = [];
      if (val.constructor == Array) {
        buf.push("[");
        for (var index = 0; index < val.length; index++) {
          buf.push(index > 0 ? propsep : linesep);
          buf.push(
            tab, prettyPrintJSON(val[index], indent, linesep, depth + 1)
          );
        }
        if (index >= 0) buf.push(linesep, tab.substr(indent));
        buf.push("]");
      } else {
        buf.push("{");
        var index = 0;
        for (var key in val) {
          if (!val.hasOwnProperty(key)) continue;
          buf.push(index > 0 ? propsep : linesep);
          buf.push(
            tab, key.toJSONString(), ": ",
            prettyPrintJSON(val[key], indent, linesep, depth + 1)
          );
          index++;
        }
        if (index >= 0) buf.push(linesep, tab.substr(indent));
        buf.push("}");
      }
      return buf.join("");
    }
  }
}
